//
//  Session+Additions.m
//  Partiture
//
//  Created by Kaarel Kruus on 16/03/14.
//  Copyright (c) 2014 Kaarel Kruus. All rights reserved.
//

#import "Session+Additions.h"
#import "PRTAppDelegate.h"
#import "PRTUtils.h"
#import "PRTCommunicationManager.h"
#import "PRTAlertView.h"
#import "Member+Additions.h"
#import "Membership+Additions.h"
#import "Membership+Enums.h"
#import "Layer+Additions.h"
#import "Participation+Additions.h"
#import "PRTPageTurnInfo.h"

@implementation Session (Additions)

+ (Session *)createAndOpenPublished:(BOOL)isPublished inRoom:(Room *)room {
    
    Session *newSession = [NSEntityDescription insertNewObjectForEntityForName:NSStringFromClass(self.class)
                                                        inManagedObjectContext:[APPDELEGATE managedObjectContext]];
    newSession.uuid = [PRTUtils generateUUID];
    newSession.room = room;
    newSession.startDate = [NSDate date];
    newSession.isPublished = isPublished;
    
    // each session should always have 1 public and 1 private layer
    Layer *private = [Layer privateLayer];
    [newSession addLayersObject:private];
    
    Layer *public = [Layer publicLayer];
    [newSession addLayersObject:public];
    
    [APPDELEGATE saveContext];
    
    // each session should always have at least 1 participant (the owner itself)
    Participation *participation;
    participation = [Participation participationForMember:[Member fetchMemberWithToken:[APPDELEGATE appUser].token]
                                                inSession:newSession];
    
    PRT_DB_CREATE_INSTANCE_LOG(NSStringFromClass(self.class), nil);
    
    return newSession;
}

+ (Session *)createAndStartUnpublishedInRoom:(Room *)room {
    PRT_SESSION_LOG(@"%s", __PRETTY_FUNCTION__);
    return [self createAndOpenPublished:NO inRoom:room];
}

+ (Session *)createAndStartPublishedInRoom:(Room *)room {
    PRT_SESSION_LOG(@"%s", __PRETTY_FUNCTION__);
    Session *sharedSession = [self createAndOpenPublished:YES inRoom:room];
    if(sharedSession) {
        [COM_PUBLISHER publishSession:sharedSession];
    }
    return sharedSession;
}

- (void)saveAndEnd {
    PRT_SESSION_LOG(@"%s", __PRETTY_FUNCTION__);
    
    NSDate *endDate = [NSDate date];
    
    self.endDate = endDate;
    
    Member *member = [Member fetchMemberWithToken:[APPDELEGATE appUser].token];
    Membership *membership = [self.room fetchMemberShipOfMember:member];
    if([membership.role isEqual:@(MembershipRoleRoomOwner)]) {
        
        Participation *participation = [self participationForMember:member];
        participation.endDate = endDate;
        
        
    } else {
        // TODO: what here?
    }
    
    if(self.isPublished) {
        // TODO: send "bye" to all connected members
        [COM_PUBLISHER unpublishSession];
    }
    [APPDELEGATE saveContext];
}

- (Layer *)fetchPublicLayer {
    return [self fetchLayer:YES];
}

- (Layer *)fetchPrivateLayer {
    return [self fetchLayer:NO];
}

- (Layer *)fetchLayer:(BOOL)isPublic {
    for (Layer *layer in self.layers.allObjects) {
        if(layer.isPublic == isPublic) {
            return layer;
        }
    }
    return nil;
}

- (Participation *)participationForMember:(Member *)member {
    for (Participation *participation in self.participations.allObjects) {
        if([participation.member isEqual:member]) {
            return participation;
        }
    }
    return nil;
}

- (PRTParticipantAddStatus)addParticipant:(Member *)member {
    Membership *membership = [self.room fetchMemberShipOfMember:member];
    if(membership) { // only a accepted member can be added as the room participant
        NSArray *activeParticipations = [member fetchActiveParticipations];
        if(activeParticipations.count > 0) { // check if there is already a started participation
            return PRTParticipantAddStatusExisting;
        } else {
            Participation *participation;
            participation = [Participation participationForMember:member inSession:self];
            return PRTParticipantAddStatusAdded;
        }
    }
    return PRTParticipantAddStatusUnknownMember;
}

- (PRTParticipantRemoveStatus)removeParticipant:(Member *)member {
    // TODO
    return PRTParticipantRemoveStatusUnknown;
}

@end

@implementation Session (COMPublisher)

- (void)pageTurn:(NSInteger)toIndex {
    PRT_SESSION_LOG(@"%s (%i)", __PRETTY_FUNCTION__, toIndex);
    
    PRTPageTurnInfo *info = [PRTPageTurnInfo infoWithPageIndex:toIndex];
    [COM_PUBLISHER pushPageTurn:info];
}

- (void)receiveJoinRequestFromUserWithName:(NSString *)name token:(NSString *)token andPasskey:(NSString *)passkey withDecisionBlock:(void(^)(BOOL accept))decisionBlock {
    PRT_SESSION_LOG(@"%s", __PRETTY_FUNCTION__);
    
    if([passkey isEqualToString:self.room.passkey]) { // TODO: what if the passkey was changed at some point?
        
        // the member might exists, but not a membership relation
        __block Member *member = [Member fetchMemberWithToken:token];
        __block Membership *membership = [Membership fetchMembershipForMember:member inRoom:self.room];
        
        if(!membership ||
           ([membership.status isEqual:@(MembershipStatusAccepted)]) ||
           ([membership.status isEqual:@(MembershipStatusUnknown)])) {
            
            PRTAlertView *alert = [[PRTAlertView alloc] init];
            
            alert.title = NSLocalizedString(@"Join.Request.Title", nil);
            alert.message = [NSString stringWithFormat:NSLocalizedString(@"Join.Request.Message", nil), name];
            
            [alert addButtonWithTitle:NSLocalizedString(@"Join.Request.Accept", nil)];
            [alert addButtonWithTitle:NSLocalizedString(@"Join.Request.AcceptPermanently", nil)];
            [alert addButtonWithTitle:NSLocalizedString(@"Join.Request.Block", nil)];
            
            [alert setTappedButtonBlock:^(NSInteger buttonIndex) {
                
                if(!member) {
                    PRT_FEATURE_ON_JOIN_REQUEST_LOG(@"No member for this token exists, create one");
                    member = [Member memberWithName:name andToken:token];
                }
                
                if(!membership) {
                    PRT_FEATURE_ON_JOIN_REQUEST_LOG(@"No membership between the member and room exists, create one");
                    membership = [Membership membershipWithMember:member inRoom:self.room];
                }
                
                if(buttonIndex == 0) {
                    membership.status = @(MembershipStatusAccepted);
                } else if(buttonIndex == 1) {
                    membership.status = @(MembershipStatusAcceptedPermanently);
                } else if(buttonIndex == 2) {
                    membership.status = @(MembershipStatusBlocked);
                }
                
                [APPDELEGATE saveContext];
                
                PRTParticipantAddStatus status = [self addParticipant:member];
                if(status == PRTParticipantAddStatusAdded) {
                    
                } else {
                    PRT_ERROR_LOG(@"%s (unexpected error - the member could not be added as an participant to the session)", __PRETTY_FUNCTION__);
                }
                
                if(buttonIndex == 0 || buttonIndex == 1) {
                    decisionBlock(YES);
                } else {
                    decisionBlock(NO);
                }
            }];
            
            DISPATCH_ASYNC_MAIN({
                [alert show];
            });
            
        } else if([membership.status isEqual:@(MembershipStatusBlocked)]) {
            PRT_FEATURE_ON_JOIN_REQUEST_LOG(@"This member has been blocked, reject request without propmting");
            decisionBlock(NO);
        } else {
            PRT_FEATURE_ON_JOIN_REQUEST_LOG(@"This member has been accepted permanently, prompting not needed");
            decisionBlock(YES);
        }
    } else {
        // the submited passkey is wrong, so reject automatically
        PRT_FEATURE_ON_JOIN_REQUEST_LOG(@"Incorrect passkey submitted, reject request without propmting");
        decisionBlock(NO);
    }
}

@end

@implementation Session (COMListener)



@end
